------------------
-- Field module --
------------------
module Field where

import Data.Char
import System.Random

-------------------------------------------------------------------------------------------
-- Empty field and cell operations
-------------------------------------------------------------------------------------------

-------------------------------------------------------------------------------------------
-- Data types 
-------------------------------------------------------------------------------------------
data Cell
	= Closed
	| Open Int
	| Flagged
	| Mined
	deriving(Eq, Show)
	
type Coord
	= (Int, Int)
	
type Field
	= [(Coord,Cell)]



-------------------------------------------------------------------------------------------	
-- Functions
-------------------------------------------------------------------------------------------

-- height 
-- width
-- amount of mines
emptyField :: Int -> Int -> Int -> Field
-- emptyField h w m
emptyField w h m
	= generateEmptyField h w m 0 0
generateEmptyField h w m hPos wPos
	| hPos == 0 && wPos == w = []
	| hPos < h = ((hPos, wPos), Closed) : generateEmptyField h w m (hPos+1) wPos
	| wPos < w = generateEmptyField h w m 0 (wPos+1)
	
getCell :: Field -> Coord -> Maybe Cell
getCell (item:restField) coord
	| coord == fst(item) = Just (snd(item))
	| otherwise = getCell restField coord 
getCell _ coord
	= Nothing

determineCell :: Maybe Cell -> Cell
determineCell (Just a) = a
determineCell Nothing = error "Cell not found"

surroundingCoords :: Coord -> [Coord]
surroundingCoords coord
	= [(x-1, y-1), (x, y-1), (x+1, y-1), (x-1, y), (x+1, y), (x-1, y+1), (x, y+1), (x+1, y+1)]
	where
		x = fst(coord)
		y = snd(coord)

existingSurroundingCoords :: Field -> [Coord] -> [Coord]
existingSurroundingCoords _ [] = []
existingSurroundingCoords field (c:coords)
	| getCell field c == Nothing = existingSurroundingCoords field coords
	| otherwise = c : existingSurroundingCoords field coords


surroundingCells :: Field -> Coord -> [(Coord, Cell)]
surroundingCells field coord
	= [(coordinate, determineCell(getCell field coordinate)) | coordinate <- existingSurroundingCoords field (surroundingCoords(coord))]
	
showField :: Field -> String
showField (item:tail)
	| x == 0 = "\n" ++ c ++ " " ++ showField tail
	| otherwise = c ++ " " ++ showField tail
	where
		x = fst(fst(item))
		c = [showCell(snd(item))]
showField _ = []

printField :: Field -> IO()
printField field
	= putStrLn(fieldString)
	where
		fieldString = showField(field)

showCell :: Cell -> Char
showCell Closed
	= '.';
showCell Flagged
	= 'x';
showCell (Open x)
	= intToDigit(x)
showCell Mined
	= 'm'

getOpenCellInt :: Cell -> Int
getOpenCellInt (Open n) = n;
getOpenCellInt _ = 0;

update :: [(Coord, Cell)] -> Field -> Field
update (newCell:restOfList) field
	= update restOfList (updateOneCell newCell field)
update [] field
	= field

updateOneCell :: (Coord, Cell) -> Field -> Field
updateOneCell (coord, cell) (item:restOfField) 
	| coord == fst(item) = (coord,cell) : updateOneCell (coord, cell) restOfField
	| otherwise = item : updateOneCell (coord, cell) restOfField
updateOneCell _ []
	= []



-------------------------------------------------------------------------------------------
-- Mined field generator
-------------------------------------------------------------------------------------------

diff_select' :: Int -> [Cell] -> [Int] -> [Cell] 
diff_select' 0 _ _ = []
diff_select' _ [] _ = error "too few elements to choose from"
diff_select' n xs (seed:rest) =
	(xs!!seed) : diff_select' (n-1) (take seed xs ++ drop (seed+1) xs) rest

randomPermutation :: [Cell] -> Int -> [Cell]
randomPermutation xs seed
	= diff_select' (length xs) xs (randList (length xs) 0 ((length xs)-1) seed)

-- height
-- width
-- amount of mines
myRandomField :: Int -> Int -> Int -> Int -> Field
myRandomField h w m seed
	= setMinesOntoField field (randomPermutation (makeList Closed (h*w-m) ++ makeList Mined m) seed)
	where
		field = emptyField h w m 

setMinesOntoField :: Field -> [Cell] -> Field
setMinesOntoField (item:restOfField) (cell:restOfCells)
	= (fst(item), cell) : setMinesOntoField restOfField restOfCells
setMinesOntoField _ _
	= []

makeList :: Cell -> Int -> [Cell]
makeList cell count
	| count == 0 = []
	| otherwise = cell : makeList cell (count-1)

randList :: Int -> Int -> Int -> Int -> [Int]
randList 0 _ _ _ = []
randList amount from to seed =
	take 1 (randomRs (from::Int,to) (mkStdGen seed)) ++ randList (amount-1) from (to-1) seed



-------------------------------------------------------------------------------------------
-- Field type casting from Field to Rand [[Int]] to match signature
-------------------------------------------------------------------------------------------

type Rand a = StdGen -> (a, StdGen)

returnR :: a -> Rand a
returnR x = \gen -> (x, gen)

randomFieldToInt :: Field -> Int -> Int -> Int -> [[Int]]
randomFieldToInt _ 0 _ _ = []
randomFieldToInt field row h w =
	getRow field (h-row) w : randomFieldToInt field (row-1) h w
		

getRow :: Field -> Int -> Int -> [Int]
getRow (item:rest) row width
	| (snd(fst(item)) == row) && (fst(fst(item)) < width) = cellToInt (snd(item)) : getRow rest row width
	| otherwise = getRow rest row width
getRow [] _ _ = []

cellToInt :: Cell -> Int
cellToInt Closed
	= 0;
cellToInt Mined
	= 1

randomField :: Int -> Int -> Int -> Int -> Rand [[Int]]
randomField h w m seed =
	returnR (randomFieldToInt (myRandomField h w m seed) h h w)


-------------------
-- End of module -- 
-------------------



-------------
-- Sandbox --
-------------

main = do
	n <- randomRIO(1::Int,1000::Int)
	printField(myRandomField 5 5 10 n)

--------------------
-- End of sandbox --
--------------------