module Swiss (Game, tournament) where

import ExcludeSet
import ThreadUtil

import Control.Monad.List
import Control.Concurrent.MVar
import Control.Concurrent
import Control.Exception (finally)
import Control.Arrow (second)
import Data.List (sortBy, groupBy)

on :: (b -> b -> c) -> (a -> b) -> a -> a -> c
(*) `on` f = \x y -> f x * f y

picks :: [a] -> [(a, [a])]
picks [] = []
picks (x : xs) = (x, xs) : map (second (x:)) (picks xs)

pair :: Ord a => ExcludeSet a -> [a] -> [[(a, a)]]
pair exs [] = return []
pair exs (x : xs) = do
  (y, ys) <- picks xs
  guard (not (areExcluded x y exs))
  liftM ((x, y) :) (pair exs ys)

-- one of the players might be a bye, it
-- loses every game played
data Scored a
  = Bye
  | Scored {
    name :: a,
    wins :: Int,
    ties :: Int
  }
  deriving (Show)

instance Eq a => Eq (Scored a) where
  Bye == Bye = True
  Scored p _ _ == Scored p' _ _ = p == p'
  _ == _ = False

instance Ord a => Ord (Scored a) where
  compare Bye Bye = EQ
  compare Bye (Scored _ _ _) = LT
  compare (Scored _ _ _) Bye = GT
  compare (Scored p _ _) (Scored p' _ _) = compare p p'

value :: Scored a -> Int
value Bye = -1
value s = 2 * wins s + ties s

mkScored :: a -> Scored a
mkScored x = Scored { name = x, wins = 0, ties = 0 } 

isBye :: Scored a -> Bool
isBye Bye = True
isBye _ = False

removeBye :: [Scored a] -> [Scored a]
removeBye = filter (not . isBye)

addWin :: Scored a -> Scored a
addWin s = s { wins = wins s + 1 }

addTie :: Scored a -> Scored a
addTie s = s { ties = ties s + 1 }

sortScored :: [Scored a] -> [Scored a]
sortScored = sortBy (revCompare `on` value)
  where
    revCompare l r = compare r l

----
-- state of our little tournament system
data SwissState a = SwissState {
    players :: [Scored a],
    excluded :: ExcludeSet (Scored a)
  }
  deriving Show

match :: Ord a => SwissState a -> [(Scored a, Scored a)]
match st = case pair (excluded st) (players st) of
  match : _ -> match
  [] -> error $ "Bug in implementation (match in swiss.hs)!"

mkSwissState :: Ord a => [a] -> SwissState a
mkSwissState xs = SwissState {
    players = if odd (length xs)
      then map mkScored xs ++ [Bye]
      else map mkScored xs,
    excluded = empty
  }

getOrder :: SwissState a -> [[a]]
getOrder = map (map name) . groupBy ((==) `on` value) . removeBye . players

type Game a = a -> a -> IO Ordering

type LiftedGame a = Scored a -> Scored a -> IO Ordering

liftGame :: Game a -> LiftedGame a
liftGame game Bye Bye = return EQ -- erp
liftGame game Bye _   = return LT
liftGame game _   Bye = return GT
liftGame game (Scored l _ _) (Scored r _ _) = game l r

-- following is still little ugly
-- doesn't make very clever stuff if client bugs out
tournament :: Ord a => Game a -> [a] -> IO [[a]]
tournament game xs = loop n (mkSwissState xs)
  where
    -- number of rounds
    n = ceiling . logBase 2 . fromIntegral $ length xs

    updateScore (l, r) LT = (l, addWin r)
    updateScore (l, r) EQ = (addTie l, addTie r)
    updateScore (l, r) GT = (addWin l, r)

    fixScores ps = sortScored (foldr (\(l, r) -> (l:) . (r:)) [] ps)

    game' = liftGame game
    
    -- what we do each round is:
    -- 1. find matches for players, this is possible as we
    --    have even number of participants with possible bye
    -- 2. play games in parallel
    -- 3. update scores and state of system
    loop 0 st = return (getOrder st)
    loop n st = do
      children <- newMVar []
      let
	ps = match st
	play (l, r) = do
	  mv <- newEmptyMVar 
	  forkChild children $ do
	    v <- game' l r
	    putMVar mv v
	  return mv 
      mvs <- mapM play ps
      waitForChildren children
      vs <- sequence `fmap` mapM tryTakeMVar mvs
      case vs of
	Nothing -> error "Did \'Game\' function throw exception?"
	Just vs -> loop (n - 1) $ st {
	    players = fixScores (zipWith updateScore ps vs),
	    excluded = foldr (\(l, r) -> exclude l r) (excluded st) ps
	  }
